/* -*- Mode: C; indent-tabs-mode: nil; c-basic-offset: 8-*- */

/*
 *This file is part of MlView.
 *
 *MlView is free software; you can redistribute it 
 *and/or modify it under the terms of 
 *the GNU General Public License as published 
 *by the Free Software Foundation; either version 2, 
 *or (at your option) any later version.
 *
 *MlView is distributed in the hope that it will be useful, 
 *but WITHOUT ANY WARRANTY; 
 *without even the implied warranty of 
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *See the GNU General Public License for more details.
 *
 *You should have received a copy of the 
 *GNU General Public License along with MlView; 
 *see the file COPYING. If not, write to the Free Software Foundation, Inc., 
 *59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *See COPYRIGHT file for copyright information.
 */

#include <string.h>
#include <libxml/uri.h>
#include <libxml/tree.h>
#include "mlview-editor.h"
#include "mlview-tree-view.h"
#include "mlview-file-descriptor.h"
#include "mlview-xml-document.h"

#include "mlview-utils.h"
#include "mlview-xslt-utils.h"


/**
 *@file
 *The definition of the methods of #MlViewEditor class.
 */

enum {
        DOCUMENT_CHANGED,
        SIGNAL_NUM
};

enum ContextualMenuEntry {
        CONTEXTUAL_MENU_ENTRY_START_BOUND,
        OPEN_FILE,
        CLOSE_FILE,
        SAVE_FILE,
        SEARCH,
        LAST_CONTEXTUAL_MENU_ENTRY_END_BOUND
};

/*======================================================
 *Some private (static) data structures used by
 *MlViewEditor widget.
 *=====================================================*/

/**
 *The #MlViewEditor private data member structure.
 *The fields of this structure must be accessed only by 
 *the mean of the available accessors.
 */
struct _MlViewEditorPrivate {
        /**
	 *The set of instances of #MlViewXMLDocument
	 *opened in the editor. The key is the pointer to 
	 *the instance of #MlViewXMLDocument opened in the editor
	 *The value is a an hash table that contains the 
	 *instances of #MlViewXMLDocumentView that holds 
	 *the key #MlViewXMLDocument.
	 */
        GHashTable *mlview_xml_docs;

        /**
	 *A hash table where the keys are the instances of
	 *MlViewXMLDocumentViews and the values are their
	 *associated document.
	 */
        GHashTable *mlview_xml_doc_views;

        /**
	 *The notebook that holds all the notebook pages (document views)
	 */
        GtkNotebook *notebook;

        /**
	 *The current document being viewed by the user
	 */
        MlViewIView *cur_view ;

        /**
	 *a hash table which keys are the base name of 
	 *the files already opened. 
	 *The associated data is the number 
	 *of times the file name
	 *has been opened. When destroying this hash 
	 *table, do not destroy the referenced data (base names) because
	 *the base names are substring of file_paths. 
	 *And file names are hold by the instances of MlViewXMLDocument.
	 */
        GHashTable *opened_file_base_names;

        /**
	 *a hash table wich keys are the file paths of 
	 *the file files already opened.
	 *the associated MlViewXMLDocument. 
	 *When destroying this hashtable do not 
	 *destroy the referenced data (file path)
	 *because file paths are hold by instances 
	 *of MlViewXMLDocument and destroyed by them.
	 */
        GHashTable *opened_file_paths;

        /**
	 *An hash table that associates the opened document label names to
	 *the matching opened view.
	 */
        GHashTable *opened_document_label_names;

        /**
         *Number of untitled document opened.
         */
        guint untitled_docs_num;

        /**
         *total number of docs opened
         */
        guint opened_docs_num;

        /**
         *The context of the application
         */
        MlViewAppContext *app_context;

        /*
         *the editor contextual menu 
         */
        GtkWidget *contextual_menu;

        gboolean dispose_has_run ;
};

#define PRIVATE(editor) (editor->priv)

static void mlview_editor_class_init (MlViewEditorClass * a_klass);

static void mlview_editor_init (MlViewEditor * a_this);

/*signal callbacks and default handlers*/
static void mlview_editor_switch_notebook_page_cb (GtkNotebook * a_notebook,
                                                   GtkNotebookPage * a_page,
                                                   gint a_page_num,
                                                   MlViewEditor * a_this);

static gboolean mlview_editor_event_cb (GtkWidget * a_widget,
                                        GdkEvent * a_event,
                                        gpointer a_user_data);

static void menu_open_local_file_activate_cb (GtkMenuItem * a_menu_item,
                                              gpointer * a_user_data);

static void menu_close_file_activate_cb (GtkMenuItem * a_menu_item,
                                         gpointer * a_user_data);

static void menu_save_file_activate_cb (GtkMenuItem * a_menu_item,
                                        gpointer * a_user_data);

static void menu_search_activate_cb (GtkMenuItem * a_menu_item,
                                     gpointer * a_user_data);

static void menu_new_document_activate_cb (GtkMenuItem * a_menu_item,
                                           gpointer * a_user_data);

static void view_name_changed_cb (MlViewIView * a_view,
                                  gpointer a_this);

static GList *build_view_list_from_hashtable (GHashTable *a_views);

static GList *build_doc_list_from_hashtable (GHashTable * a_docs);

static void add_hash_key_to_list (gpointer a_key,
                                  gpointer a_value, GList ** a_list);

static void mlview_editor_finalize (GObject *a_this) ;

static void mlview_editor_dispose (GObject *a_this) ;

static GtkVBoxClass *gv_parent_class = NULL;
static guint mlview_editor_signals[SIGNAL_NUM] = { 0 };



/*============================================================
 *private method required by the GTK typing system
 *=============================================================*/


/**
 *The vtable initialyzer.
 *@param a_klass the vtable structure.
 *
 */
static void
mlview_editor_class_init (MlViewEditorClass * a_klass)
{
        GObjectClass *object_class = NULL ;

        gv_parent_class = g_type_class_peek_parent (a_klass);

        object_class = G_OBJECT_CLASS (a_klass) ;
        g_return_if_fail (object_class) ;

        /*overload the destroy method of GtkObject */
        object_class->dispose = mlview_editor_dispose ;
        object_class->finalize = mlview_editor_finalize ;

        /*define the signals */
        mlview_editor_signals[DOCUMENT_CHANGED] =
                g_signal_new ("document-changed",
                              G_TYPE_FROM_CLASS (object_class),
                              GTK_RUN_FIRST,
                              GTK_SIGNAL_OFFSET
                              (MlViewEditorClass,
                               document_changed), NULL, NULL,
                              gtk_marshal_NONE__NONE,
                              GTK_TYPE_NONE, 0, NULL);

        /*gtk_object_class_add_signals (object_class, mlview_editor_signals, 
           SIGNAL_NUM) ; */
        a_klass->document_changed = NULL;
}


/**
 *The #MlViewEditor instance allocator.
 *Allocates the space necessary to the members of the #MlViewEditor structure
 *and does very basic initialyzation.
 *
 *@param a_this the instance of #MlViewEditor to allocate.
 */
static void
mlview_editor_init (MlViewEditor * a_this)
{
        g_assert (a_this != NULL);

        PRIVATE (a_this) =
                g_malloc0 (sizeof (MlViewEditorPrivate));

        PRIVATE (a_this)->notebook =
                GTK_NOTEBOOK (gtk_notebook_new ());

        g_signal_connect (G_OBJECT
                          (PRIVATE (a_this)->notebook),
                          "switch-page",
                          G_CALLBACK
                          (mlview_editor_switch_notebook_page_cb),
                          a_this);

        gtk_box_pack_start (GTK_BOX (a_this),
                            GTK_WIDGET (PRIVATE (a_this)->
                                        notebook), TRUE, TRUE,
                            0);

        PRIVATE (a_this)->opened_file_base_names =
                g_hash_table_new (g_str_hash, g_str_equal);

        PRIVATE (a_this)->opened_file_paths =
                g_hash_table_new (g_str_hash, g_str_equal);

        PRIVATE (a_this)->opened_document_label_names =
                g_hash_table_new (g_str_hash, g_str_equal);

        PRIVATE (a_this)->mlview_xml_docs =
                g_hash_table_new (g_direct_hash, g_direct_equal);

        PRIVATE (a_this)->mlview_xml_doc_views =
                g_hash_table_new (g_direct_hash, g_direct_equal);

        PRIVATE (a_this)->untitled_docs_num = 0;

        PRIVATE (a_this)->opened_docs_num = 0;
}


static void
mlview_editor_dispose (GObject *a_this)
{
        MlViewEditor *editor = NULL ;

        g_return_if_fail (a_this && MLVIEW_IS_EDITOR (a_this)) ;
        editor = MLVIEW_EDITOR (a_this) ;
        g_return_if_fail (PRIVATE (editor))  ;

        if (PRIVATE (editor)->dispose_has_run == TRUE)
                return ;
        /*
         *destroy each remaining view.
         **/
        if (PRIVATE (editor)->mlview_xml_doc_views) {
                GList *list_item = NULL ;
                GList *view_list = build_view_list_from_hashtable 
                        (PRIVATE (editor)->mlview_xml_doc_views) ;

                for (list_item = view_list ;
                     list_item ;list_item = list_item->next) {
                        if (list_item->data 
                            && MLVIEW_IS_IVIEW (list_item->data)) {
                                mlview_editor_remove_xml_document_view 
                                        (editor, MLVIEW_IVIEW 
                                         (list_item->data)) ;
                        }
                }
                g_hash_table_destroy
                        (PRIVATE (editor)->mlview_xml_doc_views);
                PRIVATE (editor)->mlview_xml_doc_views = NULL;
        }
        /*
         *do not free this element because 
         *it has been freed as an element of the link list above
         */
        PRIVATE (editor)->cur_view = NULL;

        if (PRIVATE (editor)->mlview_xml_docs) {
                g_hash_table_destroy (PRIVATE (editor)->mlview_xml_docs) ;
                PRIVATE (editor)->mlview_xml_docs = NULL ;
        }
        if (PRIVATE (editor)->opened_file_base_names) {
                g_hash_table_destroy
                        (PRIVATE (editor)->
                         opened_file_base_names);

                PRIVATE (editor)->opened_file_base_names = NULL;
        }
        if (PRIVATE (editor)->opened_file_paths) {
                g_hash_table_destroy
                        (PRIVATE (editor)->opened_file_paths);

                PRIVATE (editor)->opened_file_paths = NULL;
        }

        if (PRIVATE (editor)->opened_document_label_names) {
                g_hash_table_destroy
                        (PRIVATE (editor)->
                         opened_document_label_names);

                PRIVATE (editor)->opened_document_label_names = NULL;
        }

        if (G_OBJECT_CLASS (gv_parent_class)->dispose) {
                G_OBJECT_CLASS (gv_parent_class)->dispose (a_this) ;
        }
}


static void
mlview_editor_finalize (GObject *a_this)
{
        MlViewEditor *editor = NULL ;

        g_return_if_fail (a_this && MLVIEW_IS_EDITOR (a_this)) ;
        
        editor = MLVIEW_EDITOR (a_this) ;

        if (PRIVATE (editor)) {
                g_free (PRIVATE (editor)) ;
                PRIVATE (editor) = NULL ;
        }
}

/*================================================
 *private helper functions
 *
 *=================================================*/

/**
 *Builds the contextual menu that appears when the user
 *right clicks on mlview.
 *
 *@param a_this the current instance of #MlViewEditor.
 */
static void
mlview_editor_build_contextual_menu (MlViewEditor * a_this)
{
        GtkWidget *menu_item = NULL;

        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);

        if (PRIVATE (a_this)->contextual_menu) {
                gtk_widget_destroy (GTK_WIDGET
                                    (PRIVATE (a_this)->
                                     contextual_menu));
                PRIVATE (a_this)->contextual_menu = NULL;
        }

        PRIVATE (a_this)->contextual_menu  = gtk_menu_new ();

        menu_item =
                gtk_menu_item_new_with_label (_("Open File"));

        g_signal_connect (GTK_OBJECT (menu_item), "activate",
                          G_CALLBACK
                          (menu_open_local_file_activate_cb),
                          a_this);
        /* gtk_menu_popup ((PRIVATE (a_this)->contextual_menu),
           NULL,
           NULL,
           NULL,
           GDK_BUTTON_PRESSED,
           NULL); */

        gtk_menu_append (GTK_MENU
                         (PRIVATE (a_this)->contextual_menu),
                         menu_item);

        gtk_widget_show (menu_item);

        menu_item =
                gtk_menu_item_new_with_label (_("Close File"));

        g_signal_connect (G_OBJECT (menu_item), "activate",
                          G_CALLBACK
                          (menu_close_file_activate_cb),
                          a_this);

        gtk_menu_append (GTK_MENU
                         (PRIVATE (a_this)->contextual_menu),
                         menu_item);

        gtk_widget_show (menu_item);

        menu_item =
                gtk_menu_item_new_with_label (_("Save File"));

        g_signal_connect (G_OBJECT (menu_item), "activate",
                          G_CALLBACK
                          (menu_save_file_activate_cb),
                          a_this);

        gtk_menu_append (GTK_MENU
                         (PRIVATE (a_this)->contextual_menu),
                         menu_item);

        gtk_widget_show (menu_item);

        /*separator */
        menu_item = gtk_menu_item_new ();

        gtk_menu_append (GTK_MENU
                         (PRIVATE (a_this)->contextual_menu),
                         menu_item);

        gtk_widget_show (menu_item);

        menu_item = gtk_menu_item_new_with_label (_("Search"));

        g_signal_connect (G_OBJECT (menu_item), "activate",
                          G_CALLBACK (menu_search_activate_cb),
                          a_this);

        gtk_menu_append (GTK_MENU
                         (PRIVATE (a_this)->contextual_menu),
                         menu_item);

        gtk_widget_show (menu_item);

        /*separator */
        menu_item = gtk_menu_item_new ();

        gtk_menu_append (GTK_MENU
                         (PRIVATE (a_this)->contextual_menu),
                         menu_item);

        gtk_widget_show (menu_item);

        menu_item =
                gtk_menu_item_new_with_label (_("New Document"));

        gtk_signal_connect (GTK_OBJECT (menu_item), "activate",
                            GTK_SIGNAL_FUNC
                            (menu_new_document_activate_cb),
                            a_this);

        gtk_menu_append (GTK_MENU
                         (PRIVATE (a_this)->contextual_menu),
                         menu_item);

        gtk_widget_show (menu_item);

        if (PRIVATE (a_this)->cur_view) {
                mlview_iview_update_contextual_menu
                        (MLVIEW_IVIEW (PRIVATE (a_this)->cur_view),
                         &PRIVATE (a_this)->contextual_menu) ;
        }
}

/*================================================
 *signal callbacks and default handlers
 *
 *=================================================*/

/**
 *this callback is called when the editor
 *switch from one view to another.
 *
 *@param a_notebook the notebook editor that holds the view notebook pages.
 *@param a_page the notebook page that holds the view 
 *(the view is an instance of #MlViewXMLDocumentView)
 *@param a_page_num the number of the new notebook page.
 *@param a_this the current instance of #MlViewEditor.
 */
static void
mlview_editor_switch_notebook_page_cb (GtkNotebook * a_notebook,
                                       GtkNotebookPage * a_page,
                                       gint a_page_num,
                                       MlViewEditor * a_this)
{
        MlViewIView *doc_view = NULL ;
        GtkWidget *cur_child_widget = NULL;

        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);

        cur_child_widget = gtk_notebook_get_nth_page (a_notebook,
                                                      a_page_num);
        g_return_if_fail (cur_child_widget);
        doc_view = MLVIEW_IVIEW (cur_child_widget) ;
        g_return_if_fail (doc_view != NULL);

        PRIVATE (a_this)->cur_view = doc_view;
}

/*
 *TODO: replace all the MlViewXMLDocumentView references 
 *by MlViewIView.
 *do not forget to call mlview_iview_connect_to_doc() during
 *the instanciation of a view.
 */

/**
 *This callback is called whenever a graphical
 *even is generated in the editor's main window.
 *It is here that detect that the user has right clicked and
 *we trigger the building of the contextual menu.
 */
static gboolean
mlview_editor_event_cb (GtkWidget * a_widget,
                        GdkEvent * a_event, gpointer a_user_data)
{

        MlViewEditor *editor = NULL;

        g_return_val_if_fail (a_widget != NULL, FALSE);
        g_return_val_if_fail (GTK_IS_WIDGET (a_widget), FALSE);
        g_return_val_if_fail (a_user_data != NULL, FALSE);
        g_return_val_if_fail (MLVIEW_IS_EDITOR (a_user_data),
                              FALSE);
        g_return_val_if_fail (a_event != NULL, FALSE);

        editor = MLVIEW_EDITOR (a_user_data);
        g_return_val_if_fail (editor != NULL, FALSE);
        g_return_val_if_fail (PRIVATE (editor), FALSE);

        switch (a_event->type) {
        case GDK_BUTTON_PRESS:
                if (a_event->button.button == 3) {
                        /*user pressed the right mouse button */
                        mlview_editor_build_contextual_menu
                                (editor);

                        g_return_val_if_fail
                                (PRIVATE (editor)->
                                 contextual_menu != NULL, FALSE);

                        gtk_menu_popup (GTK_MENU (PRIVATE 
                                                  (editor)->contextual_menu), 
                                        NULL,
                                        NULL, NULL, editor,
                                        a_event->button.button,
                                        a_event->button.time);
                }
                break;
        default:
                break;
        }
        return FALSE;
}

/**
 *Called when the user clicks on the
 *"open file" menuitem of the contextual menu.
 *
 *@param a_menu_item the menu item the user clicked in.
 *@param a_user_data the current instance of #MlViewEditor.
 */
static void
menu_open_local_file_activate_cb (GtkMenuItem * a_menu_item,
                                  gpointer * a_user_data)
{
        MlViewEditor *editor = NULL;

        g_return_if_fail (a_menu_item != NULL);
        g_return_if_fail (GTK_IS_MENU_ITEM (a_menu_item));
        g_return_if_fail (a_user_data != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_user_data));

        editor = MLVIEW_EDITOR (a_user_data);
        mlview_editor_open_local_xml_document_interactive
                (editor);
}


/**
 *This callback is called when the user clicks on the
 *"close file" menu item of the contextual menu.
 *
 *@param a_menu_item the menu item the user clicked in.
 *@param a_user_data the current instance of #MlViewEditor.
 */
static void
menu_close_file_activate_cb (GtkMenuItem * a_menu_item,
                             gpointer * a_user_data)
{
        MlViewEditor *editor = NULL;

        g_return_if_fail (a_menu_item != NULL);
        g_return_if_fail (GTK_IS_MENU_ITEM (a_menu_item));
        g_return_if_fail (a_user_data != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_user_data));

        editor = MLVIEW_EDITOR (a_user_data);
        mlview_editor_close_xml_document_interactive (editor);
}

/**
 *Called when the user clicks on the
 *"save as" menu item of the contextual menu.
 *
 *@param a_menu_item the menu item the used clicked in.
 *@param a_user_data the current instance of #MlViewEditor.
 */
static void
menu_save_file_activate_cb (GtkMenuItem * a_menu_item,
                            gpointer * a_user_data)
{
        MlViewEditor *editor = NULL;

        g_return_if_fail (a_menu_item != NULL);
        g_return_if_fail (GTK_IS_MENU_ITEM (a_menu_item));
        g_return_if_fail (a_user_data != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_user_data));

        editor = MLVIEW_EDITOR (a_user_data);
        mlview_editor_save_xml_document (editor);
}


/**
 *Called when the user clicks on the
 *"search" menu item of the contextual menu.
 *
 *@param a_menu_item the menu item the user clicked in.
 *@param a_user_data the current instance of #MlViewEditor.
 */
static void
menu_search_activate_cb (GtkMenuItem * a_menu_item,
                         gpointer * a_user_data)
{
        MlViewEditor *editor = NULL;
        MlViewAction action = {0} ;

        g_return_if_fail (a_menu_item != NULL);
        g_return_if_fail (GTK_IS_MENU_ITEM (a_menu_item));
        g_return_if_fail (a_user_data != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_user_data));

        editor = MLVIEW_EDITOR (a_user_data);
        action.name = (guchar*) "find-xml-node-that-contains-str-interactive" ;
        
        mlview_iview_execute_action (PRIVATE (editor)->cur_view, &action) ;

}


/**
 *Called when the user clicks on the
 *"new document" menu item of the contextual menu.
 *
 *@param a_menu_item the menu item the clicked in.
 *@param a_user_data the current instance of #MlViewEditor.
 */
static void
menu_new_document_activate_cb (GtkMenuItem * a_menu_item,
                               gpointer * a_user_data)
{
        MlViewEditor *editor = NULL;

        g_return_if_fail (a_menu_item != NULL);
        g_return_if_fail (GTK_IS_MENU_ITEM (a_menu_item));
        g_return_if_fail (a_user_data != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_user_data));

        editor = MLVIEW_EDITOR (a_user_data);
        mlview_editor_create_new_xml_document (editor);
}


/**
 *This callback is called when the name of the current
 *instance of #MlViewXMLDocumentView changes.
 *
 *@param a_view the current instance of #MlViewXMLDocumentView
 *which name has changed.
 *@param a_this the current instance of #MlViewEditor.
 */
static void
view_name_changed_cb (MlViewIView * a_view,
                      gpointer a_this)
{
        MlViewEditor *editor = NULL;
        guchar *new_view_name = NULL;

        g_return_if_fail (a_view 
                          && MLVIEW_IS_IVIEW (a_view)
                          && a_this
                          && MLVIEW_IS_EDITOR (a_this));

        editor = MLVIEW_EDITOR (a_this);
        g_return_if_fail (PRIVATE (editor) != NULL);
        g_return_if_fail (PRIVATE (editor)->notebook != NULL);

        mlview_iview_get_name (a_view, &new_view_name) ;
        g_return_if_fail (new_view_name) ;
        gtk_notebook_set_tab_label_text 
                (PRIVATE (editor)->notebook,
                 GTK_WIDGET (a_view), new_view_name);

}

static GList *
build_view_list_from_hashtable (GHashTable * a_views)
{
        GList *result = NULL;

        g_hash_table_foreach (a_views,
                              (GHFunc) add_hash_key_to_list,
                              &result);
        return result;
}

static GList *
build_doc_list_from_hashtable (GHashTable * a_docs)
{
        GList *result = NULL;

        g_hash_table_foreach (a_docs,
                              (GHFunc) add_hash_key_to_list,
                              &result);
        return result;
}

static void
add_hash_key_to_list (gpointer a_key,
                      gpointer a_value, GList ** a_list)
{

        g_return_if_fail (a_list != NULL);
        *a_list = g_list_append (*a_list, a_key);
}

/*=================================================
 *public methods
 *=================================================*/

/**
 *the standard type builder of #MlViewEditor 
 *@return the type id of #MlViewEditor.
 */
GType
mlview_editor_get_type (void)
{
        static guint type = 0;

        if (!type) {
                static const GTypeInfo type_info = {
                        sizeof (MlViewEditorClass),
                        NULL, NULL,
                        (GClassInitFunc)
                                mlview_editor_class_init,
                        NULL, NULL,
                        sizeof (MlViewEditor),
                        0,
                        (GInstanceInitFunc) mlview_editor_init
                };

                type = g_type_register_static (GTK_TYPE_VBOX,
                                               "MlViewEditor",
                                               &type_info, 0);
        }
        return type;
}


/**
 *Creates and returns a new instance of #MlViewEditor.
 *
 *@param a_title the title of the editor (not used yet).
 *@param a_context the current instance of MlViewApplicationContext.
 */
GtkWidget *
mlview_editor_new (const guchar * a_title,
                   MlViewAppContext * a_context)
{
        MlViewEditor *editor = NULL;
        gpointer tmp = NULL;

        editor = g_object_new (MLVIEW_TYPE_EDITOR, NULL);

        PRIVATE (editor)->app_context = a_context;
        g_return_val_if_fail (a_context != NULL, NULL);

        tmp = mlview_app_context_get_element (a_context,
                                              "GnomeApp");
        if (tmp && GTK_IS_WIDGET (tmp)) {
                GtkWidget *parent_app = NULL;

                parent_app = GTK_WIDGET (tmp);
                gtk_widget_add_events (parent_app,
                                       GDK_BUTTON3_MOTION_MASK);
                gtk_signal_connect (GTK_OBJECT (parent_app),
                                    "button_press_event",
                                    GTK_SIGNAL_FUNC
                                    (mlview_editor_event_cb),
                                    editor);
        }

        return GTK_WIDGET (editor);
}


/**
 *The setter of the application context associated to the 
 *current mlview editor.
 *
 *@param a_this the current instance of #MlViewEditor.
 *@param a_context the new application context.
 */
void
mlview_editor_set_app_context (MlViewEditor * a_this,
                               MlViewAppContext * a_context)
{
        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);
        g_return_if_fail (a_context != NULL);
        g_return_if_fail (MLVIEW_IS_APP_CONTEXT (a_context));

        PRIVATE (a_this)->app_context = a_context;
}


/**
 *The getter of the application context associated
 *to the current mlview editor.
 *
 *@param a_this the current instance of #MlViewEditor.
 *@return the application context associated to the editor.
 */
MlViewAppContext *
mlview_editor_get_app_context (MlViewEditor * a_this)
{
        g_return_val_if_fail (a_this != NULL, NULL);
        g_return_val_if_fail (MLVIEW_IS_EDITOR (a_this), NULL);
        g_return_val_if_fail (PRIVATE (a_this) != NULL, NULL);

        return PRIVATE (a_this)->app_context;
}


/**
 *Opens the xml file a_file_path, loads it, build a view to edit it and
 *adds this new view to the editor.
 *
 *@param a_this the current instance of #MlViewEditor.
 *@param a_file_path the path of the file to open.
 */
void
mlview_editor_load_xml_file (MlViewEditor * a_this,
                             gchar * a_file_path)
{
        MlViewXMLDocument *mlview_xml_document = NULL;

        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this));

        if (a_file_path == NULL)
                return;

        mlview_app_context_sbar_push_message (PRIVATE
                                              (a_this)->
                                              app_context,
                                              _
                                              ("Opening file %s..."),
                                              a_file_path);

        mlview_xml_document =
                mlview_xml_document_open (a_file_path,
                                          PRIVATE (a_this)->
                                          app_context);

        if (mlview_xml_document) {
                GtkWidget *parent_window = NULL;
                MlViewIView *new_view = NULL ;

                new_view = MLVIEW_IVIEW 
                        (mlview_tree_view_new 
                         (mlview_xml_document,
                          a_file_path,
                          PRIVATE (a_this)->app_context)) ;

                g_return_if_fail (new_view != NULL);

                parent_window = gtk_widget_get_toplevel
                        (GTK_WIDGET (a_this));

                gtk_widget_add_events (parent_window,
                                       GDK_BUTTON3_MOTION_MASK);

                gtk_signal_connect (GTK_OBJECT (parent_window),
                                    "button_press_event",
                                    GTK_SIGNAL_FUNC
                                    (mlview_editor_event_cb),
                                    a_this);
                mlview_editor_add_xml_document_view
                        (a_this, new_view) ;
        }

        /*
         *FIXME: add support for exception thrown by 
         *mlview_xml_document_open in case 
         *of error and display that error => BIG DEAL
         */

        mlview_app_context_sbar_pop_message (PRIVATE (a_this)->
                                             app_context);
}


/**
 *interactively open/edit a local xml file name.
 *@param a_this the current mlview editor.
 */
void
mlview_editor_open_local_xml_document_interactive (MlViewEditor *a_this)
{
        MlViewFileSelection *file_selector;
        enum MLVIEW_SELECTED_BUTTON button;
        gchar *file_name = NULL;

        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);
        g_return_if_fail (PRIVATE (a_this)->app_context !=
                          NULL);

        file_selector =
                mlview_app_context_get_file_selector
                (PRIVATE (a_this)->app_context,
                 _("Open xml document"));

        g_return_if_fail (file_selector != NULL);

        mlview_app_context_sbar_push_message
                (PRIVATE (a_this)->app_context,
                 _("Choose the xml file to open"));

        button = mlview_file_selection_run
                (MLVIEW_FILE_SELECTION (file_selector), TRUE);

        switch (button) {
        case OK_BUTTON:
                file_name =
                        g_strdup (gtk_file_selection_get_filename
                                  (GTK_FILE_SELECTION
                                   (file_selector)));

                if (file_name && strcmp (file_name, ""))
                        mlview_editor_load_xml_file (a_this,
                                                     file_name);

                if (file_name) {
                        g_free (file_name);
                        file_name = NULL;
                }
        case CANCEL_BUTTON:
        case WINDOW_CLOSED:
        default:
                break;
        }

        mlview_app_context_sbar_pop_message
                (PRIVATE (a_this)->app_context);
}


/**
 *Creates a new view on an existing document.
 *
 *@param a_this the current mlview editor.
 *@param a_xml_doc the xml document on which the new
 *view must be created.
 *@param a_view_type the type of view.
 *@return the newly created view.
 */
MlViewIView *
mlview_editor_create_new_view_on_document (MlViewEditor *a_this,
                                           MlViewXMLDocument *a_xml_doc, 
                                           enum MLVIEW_XML_DOC_VIEW_TYPE a_view_type)
{
        MlViewIView *result = NULL;

        g_return_val_if_fail (a_this != NULL, NULL);
        g_return_val_if_fail (MLVIEW_IS_EDITOR (a_this), NULL);
        g_return_val_if_fail (PRIVATE (a_this) != NULL, NULL);
        g_return_val_if_fail (a_xml_doc != NULL, NULL);
        g_return_val_if_fail (MLVIEW_IS_XML_DOCUMENT (a_xml_doc),
                              NULL);

        switch (a_view_type) {
        case MLVIEW_XML_DOC_TREE_VIEW:
                result = MLVIEW_IVIEW
                        (mlview_tree_view_new
                         (a_xml_doc, NULL,
                          PRIVATE (a_this)->app_context));

                mlview_editor_add_xml_document_view (a_this,
                                                     result);

                break;
        default:
                break;
        }

        return result;
}

/**
 *Interactively creates a new view on an existing
 *xml document.
 *The view will be created if and only if an non empty document is
 *currently selected in the editor.
 *
 *@param a_this the current mlview editor.
 *@return the newly created view or NULL if it could not be created.
 */
MlViewIView*
mlview_editor_create_new_view_on_current_document_interactive (MlViewEditor * a_this) 
{
        MlViewXMLDocument *xml_document = NULL;

        g_return_val_if_fail (a_this != NULL, NULL);
        g_return_val_if_fail (MLVIEW_IS_EDITOR (a_this), NULL);
        g_return_val_if_fail (PRIVATE (a_this) != NULL, NULL);


        if (PRIVATE (a_this)->cur_view == NULL)
                return NULL;

        mlview_iview_get_document
                (PRIVATE (a_this)->cur_view, &xml_document);

        return mlview_editor_create_new_view_on_document
                (a_this, xml_document,
                 MLVIEW_XML_DOC_TREE_VIEW);

}

/**
 *Getter of the current selected edition
 *view.
 *@param a_this the current instance of #MlViewEditor.
 *@return the current MlViewXMLDocumentView or NULL.
 */

MlViewIView *
mlview_editor_get_cur_doc_view (MlViewEditor * a_this)
{
        g_return_val_if_fail (a_this && PRIVATE (a_this), NULL);

        return PRIVATE (a_this)->cur_view;
}


/**
 *Getter of the current selected instance of
 *#MlViewXMLDocument *
 *
 *@param a_this the current instance of #MlViewEditor.
 *@return the current selected instance of #MlViewXMLDocument,
 *or NULL.
 */
MlViewXMLDocument *
mlview_editor_get_cur_doc (MlViewEditor * a_this)
{
        MlViewXMLDocument *doc = NULL ;

        g_return_val_if_fail (a_this && PRIVATE (a_this), NULL);

        if (PRIVATE (a_this)->cur_view) {
                mlview_iview_get_document
                        (PRIVATE (a_this)->cur_view, &doc);
        }

        return doc ;
}

/**
 *Getter of the glist of opened docs
 *#MlViewXMLDocument *
 *
 *@param a_this the current instance of #MlViewEditor.
 *@return a GList* of #MlViewXMLDocument, taht the caller 
 * is responsble to free
 */
GList *
mlview_editor_get_list_open_doc (MlViewEditor * a_this)
{
   GList *docs = NULL;

   docs = build_doc_list_from_hashtable
                (PRIVATE (a_this)->mlview_xml_docs);

   return docs;
}

/**
 *Setter of the name of the currently selected view.
 *
 *@param a_this the current mlview editor.
 *@param a_name the new name of the view.
 */
void
mlview_editor_set_current_view_name (MlViewEditor * a_this, gchar * a_name)
{
        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);

        if (PRIVATE (a_this)->cur_view == NULL)
                return;
        mlview_iview_set_name
                (PRIVATE (a_this)->cur_view, a_name);
}


/**
 *Interactively sets the name of the view currently
 *selected in the mlview editor.
 *
 *@param a_this the current mlview editor.
 */
void
mlview_editor_set_current_view_name_interactive (MlViewEditor *a_this)
{
        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);

        if (PRIVATE (a_this)->cur_view == NULL)
                return;

        mlview_iview_set_name_interactive
                (PRIVATE (a_this)->cur_view);
}


/**
 *Adds a document view to the editor. If a_xml_doc_view is a new 
 *document view, adds a new notebook page to the #MlViewEditor notebook.
 *The base name of the document is shown in the notebook page tab. 
 *If the underlying document of a_xml_doc_view has the 
 *same base name as a document already opened,
 *the notebook page tab string will be "basename<nb>" where 'nb' is 
 *the number of docs that have the same name in the editor (like in emacs).
 *If a_xml_doc_view is already loaded in the current instance of #MlViewEditor, 
 *the old instance of a_xml_doc is destroyed 
 *(as well as the notebook page that contains it) and the new 
 *instance of a_xml_doc is added to the editor. 
 *
 *@param a_this the current mlview editor.
 *@param a_xml_doc_view the document view to be added to the editor.
 */
void
mlview_editor_add_xml_document_view (MlViewEditor * a_this,
                                     MlViewIView *a_view)
{
        MlViewFileDescriptor *file_desc = NULL;
        guchar *file_path = NULL,
                *base_name = NULL;
        MlViewXMLDocument *mlview_xml_document = NULL;
        GHashTable *views_associated_to_document = NULL;
        MlViewIView *iview = NULL;
        gpointer ptr = NULL ;
        gboolean is_new_doc_tree = TRUE;
        GtkWidget *label = NULL, *view_impl = NULL,
                *view_impl_tmp = NULL;

        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);
        g_return_if_fail (PRIVATE (a_this)->notebook != NULL);
        g_return_if_fail (a_view != NULL);
        g_return_if_fail (MLVIEW_IS_IVIEW (a_view)) ;
        mlview_iview_get_impl (a_view, &view_impl) ;
        g_return_if_fail (view_impl) ;
        g_return_if_fail (PRIVATE (a_this)->mlview_xml_docs);
        g_return_if_fail (PRIVATE (a_this)->mlview_xml_doc_views);

        mlview_iview_get_document (a_view, &mlview_xml_document) ;
        g_return_if_fail (mlview_xml_document != NULL);

        file_desc = mlview_xml_document_get_file_descriptor 
                (mlview_xml_document) ;
        if (file_desc)
                file_path =
                        mlview_file_descriptor_get_file_path
                        (file_desc) ;
        /*check if a view on the same document has been added to
         *the editor already.
         */
        views_associated_to_document =
                g_hash_table_lookup
                (PRIVATE (a_this)->mlview_xml_docs,
                 mlview_xml_document);

        if (views_associated_to_document) {
                is_new_doc_tree = FALSE;
        }

        if (file_path == NULL) {
                gchar *tmp_str = NULL,
                        *label_str = NULL;

                if (is_new_doc_tree == TRUE)
                        PRIVATE (a_this)->untitled_docs_num++;
                tmp_str =
                        g_strdup_printf
                        ("%d",
                         PRIVATE (a_this)->untitled_docs_num);

                label_str =
                        g_strconcat ("untitled", tmp_str, NULL);
                label = gtk_label_new (label_str);
                g_free (label_str);
                g_free (tmp_str);

        } else {

                gint base_name_nb = 0;

                gboolean file_is_already_opened = FALSE;

                base_name = (guchar *) g_basename (file_path);

                if (is_new_doc_tree
                    && (iview = g_hash_table_lookup
                        (PRIVATE (a_this)->opened_file_paths,
                         file_path)) != NULL) {
                        /*
                         *There is an xml document coming from the
                         *same url that is opened already. 
                         *So, reopen it.
                         *That is, remove it previous 
                         *instance and add
                         *the new one.
                         */
                        GtkWidget *old_label = NULL ;
                        gchar *old_label_str_tmp = NULL,
                                *old_label_str = NULL;
                        
                        /*
                         *get the old label string because 
                         *this document
                         *must have the same label as 
                         *the one alreay opened
                         */
                        mlview_iview_get_impl (iview, 
                                               &view_impl_tmp) ;
                        old_label = gtk_notebook_get_tab_label
                                (PRIVATE (a_this)->notebook,
                                 view_impl_tmp);

                        g_assert (old_label != NULL);

                        gtk_label_get (GTK_LABEL (old_label),
                                       &old_label_str_tmp);

                        /*
                         *make a copy of the label string because
                         *mlview_editor_remove_xml_document ()
                         *will destroy this label.
                         */

                        /*old_label_str_tmp belongs to label,
                         *and will be freed by him
                         */
                        old_label_str =
                                g_strdup (old_label_str_tmp);

                        mlview_editor_remove_xml_document_view
                                (a_this, iview);

                        /*create the label of this document notebook page */
                        label = gtk_label_new (old_label_str);
                        /*old_label_str has been strduped */
                        g_free (old_label_str);
                        gtk_label_get (GTK_LABEL (label),
                                       &old_label_str);
                        /*
                         *old_label_str belongs to 
                         *label and will be freed by him
                         *so we can use it as a key in a hash 
                         *table without any need of
                         *freeing it.
                         */
                        g_hash_table_insert
                                (PRIVATE
                                 (a_this)->
                                 opened_document_label_names,
                                 g_strdup (old_label_str),
                                 a_view);

                        file_is_already_opened = TRUE;

                } else if ((ptr = g_hash_table_lookup
                            (PRIVATE (a_this)->
                             opened_file_base_names,
                             base_name)) == NULL) {
                        /*
                         *It is the first time a 
                         *document with this basename is opened
                         */
                        base_name_nb = 1;

                } else if (ptr != NULL) {
                        /*
                         *some documents with the this basename 
                         *are already opened
                         */
                        base_name_nb = GPOINTER_TO_INT (ptr);
                        if (!is_new_doc_tree)
                                base_name_nb++;
                }

                g_hash_table_insert
                        (PRIVATE (a_this)->opened_file_base_names, 
                         base_name, GINT_TO_POINTER (base_name_nb));
                g_hash_table_insert
                        (PRIVATE (a_this)->opened_file_paths,
                         file_path, a_view);
                if (base_name_nb > 1) {
                        gchar *tmp_str = NULL,
                                *label_str = NULL;
                        while (1) {
                                tmp_str =
                                        g_strdup_printf 
                                        ("%d", base_name_nb);
                                label_str =
                                        g_strconcat (base_name,
                                                     "<", tmp_str,
                                                     ">", NULL);

                                if (g_hash_table_lookup
                                    (PRIVATE (a_this)->
                                     opened_document_label_names,
                                     label_str)) {
                                        base_name_nb++;
                                        g_free (tmp_str);
                                        continue;
                                }
                                break;
                        }
                        mlview_iview_set_name (a_view, 
                                               label_str);
                        label = gtk_label_new (label_str);
                        g_hash_table_insert
                                (PRIVATE
                                 (a_this)->
                                 opened_document_label_names,
                                 g_strdup (label_str),
                                 a_view) ;
                        g_free (tmp_str);
                        g_free (label_str);
                } else if (file_is_already_opened == FALSE) {
                        label = gtk_label_new (base_name);
                        g_hash_table_insert
                                (PRIVATE
                                 (a_this)->
                                 opened_document_label_names,
                                 g_strdup (base_name),
                                 a_view);
                }
        }

        /*update the view->document index */
        g_hash_table_insert 
                (PRIVATE (a_this)->mlview_xml_doc_views,
                 a_view, mlview_xml_document) ;

        /*update the document->views index */
        views_associated_to_document =
                g_hash_table_lookup 
                (PRIVATE (a_this)->mlview_xml_docs,
                 mlview_xml_document);
        if (!views_associated_to_document) {
                views_associated_to_document =
                        g_hash_table_new (g_direct_hash,
                                          g_direct_equal);
                g_assert (views_associated_to_document != NULL);
                g_hash_table_insert 
                        (PRIVATE (a_this)->mlview_xml_docs,
                         mlview_xml_document,
                         views_associated_to_document);
        }
        g_hash_table_insert (views_associated_to_document,
                             a_view,
                             mlview_xml_document);
        if (is_new_doc_tree == TRUE)
                PRIVATE (a_this)->opened_docs_num++;

        /*now, visually add the view */
        gtk_notebook_append_page (PRIVATE (a_this)->notebook,
                                  view_impl, label);
        g_signal_connect (G_OBJECT (a_view),
                          "name-changed",
                          G_CALLBACK (view_name_changed_cb),
                          a_this);
        /*make sure the newly added view comes on "front"*/
        gtk_widget_show (view_impl) ;
        gtk_notebook_set_current_page
                (PRIVATE (a_this)->notebook, -1) ;
        mlview_tree_view_set_all_paned_proportions
                (MLVIEW_TREE_VIEW (a_view), 50, 70) ;
        gtk_widget_show_all (GTK_WIDGET (a_this));

}

/**
 *removes the document view a_view from the editor.
 *This method does not save the document.
 *
 *@param a_this the current mlview editor.
 *@param a_view the document view to remove from the editor.
 */
void
mlview_editor_remove_xml_document_view (MlViewEditor * a_this,
                                        MlViewIView *a_view)
{
        GtkWidget *label = NULL, *view_impl = NULL ;
        gpointer *ptr = NULL ;
        gint notebook_page_num ;
        MlViewFileDescriptor *file_desc = NULL;
        MlViewXMLDocument *mlview_xml_doc = NULL;
        GHashTable *views_related_to_document = NULL;
        gboolean doc_to_be_closed = FALSE;
        guchar *file_path = NULL,
                *base_name = NULL,
                *label_str = NULL;

        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);
        g_return_if_fail 
                (PRIVATE (a_this)->mlview_xml_doc_views != NULL);
        g_return_if_fail (a_view != NULL);
        g_return_if_fail (MLVIEW_IS_IVIEW (a_view)) ;
        mlview_iview_get_impl (a_view, &view_impl) ;
        g_return_if_fail (view_impl) ;

        mlview_iview_get_document (a_view, &mlview_xml_doc) ;
        g_return_if_fail (mlview_xml_doc);

        file_desc = mlview_xml_document_get_file_descriptor
                (mlview_xml_doc) ;
        if (file_desc)
                file_path =
                        mlview_file_descriptor_get_file_path
                        (file_desc);
        if (file_path != NULL)
                base_name = (guchar *) g_basename (file_path);       
        label = gtk_notebook_get_tab_label 
                (PRIVATE (a_this)->notebook,
                 GTK_WIDGET (a_view)) ;
        gtk_label_get (GTK_LABEL (label), (char **) &label_str);
        label_str = g_strdup (label_str);
        g_return_if_fail (label != NULL);
        /*
         *check if the document view a_view
         *is opened in this instance of MlViewEditor.
         */
        ptr = g_hash_table_lookup 
                (PRIVATE (a_this)->mlview_xml_doc_views,
                 a_view);
        g_return_if_fail (ptr != NULL);
        /*
         *removes a_view from the hashtable of 
         *the opened document views.
         *and from the opened_file_paths hash_table.
         */
        g_hash_table_remove (PRIVATE (a_this)->mlview_xml_doc_views,
                             a_view);
        views_related_to_document =
                g_hash_table_lookup 
                (PRIVATE (a_this)->mlview_xml_docs,
                 mlview_xml_doc);
        g_return_if_fail (views_related_to_document != NULL);

        ptr = g_hash_table_lookup (views_related_to_document, 
                                   a_view) ;
        g_return_if_fail (ptr != NULL) ;
        g_hash_table_remove (views_related_to_document,
                             a_view) ;
        /*removes the notebook page that contains the view */
        notebook_page_num = gtk_notebook_page_num
                (PRIVATE (a_this)->notebook, view_impl);

        g_return_if_fail (notebook_page_num != -1);

        gtk_notebook_remove_page (PRIVATE (a_this)->notebook,
                                  notebook_page_num);

        if (g_hash_table_size (views_related_to_document) == 0) {
                /*
                 *no views are opened on the current 
                 *doc anymore=>close the doc
                 */
                g_hash_table_remove (PRIVATE (a_this)->
                                     mlview_xml_docs,
                                     mlview_xml_doc);

                if (file_path != NULL) {
                        g_hash_table_remove
                                (PRIVATE (a_this)->
                                 opened_file_paths, file_path);
                }

                doc_to_be_closed = TRUE;

                PRIVATE (a_this)->opened_docs_num--;
        }

        if (doc_to_be_closed && label_str) {
                /*
                 *remove the entry in the 
                 *opened_document_label_names hash table
                 */
                g_hash_table_remove
                        (PRIVATE (a_this)->
                         opened_document_label_names, label_str);

                g_free (label_str);
                label_str = NULL;
        }

        /*if there are several docs that have the save base name as this one,
         *decrement their number, and if the number reaches 0, 
         *remove the entry matching this base name
         *from the hash table.
         */
        if (doc_to_be_closed == TRUE && file_path != NULL) {
                gint tmp_int;

                ptr = g_hash_table_lookup
                        (PRIVATE (a_this)->
                         opened_file_base_names, base_name);

                tmp_int = GPOINTER_TO_INT (ptr);
                tmp_int--;

                if (tmp_int == 0) {
                        g_hash_table_remove
                                (PRIVATE (a_this)->
                                 opened_file_base_names,
                                 base_name);
                } else {
                        ptr = GINT_TO_POINTER (tmp_int);
                        g_hash_table_insert
                                (PRIVATE (a_this)->
                                 opened_file_base_names,
                                 base_name, ptr);
                }
        } else if (doc_to_be_closed == TRUE && !file_path) {
                PRIVATE (a_this)->untitled_docs_num--;
        }

        if (doc_to_be_closed && mlview_xml_doc) {
                mlview_xml_document_unref (mlview_xml_doc);
                mlview_xml_doc = NULL;
        }
}

/**
 *Displays a dialog with an entry for the root element name.
 *It's up to the caller check the validity of the name.
 *
 *@param elname the location to return the name string.
 *@return TRUE if the user pressed OK, FALSE otherwise
 */
static gboolean
mlview_editor_ask_root_element_name (gchar **elname)
{
        GtkWidget *dialog;
        GtkWidget *hbox;
        GtkWidget *image;
        GtkWidget *label;
        GtkWidget *entry;
        gint res;
        gboolean ret = FALSE;

        g_return_val_if_fail (elname, FALSE);

        *elname = NULL;

        dialog = gtk_dialog_new_with_buttons 
                (_("Name of the root element"),
                 NULL, GTK_DIALOG_MODAL,
                 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, 
                 GTK_STOCK_OK, GTK_RESPONSE_OK,
                 NULL);

        gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
        gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);

        hbox = gtk_hbox_new (FALSE, 6);
        image = gtk_image_new_from_stock (GTK_STOCK_DIALOG_QUESTION,
                                          GTK_ICON_SIZE_DIALOG);
        gtk_misc_set_alignment (GTK_MISC (image), 0.5, 0.0);
        label = gtk_label_new (_("Root element name:"));
        entry = gtk_entry_new ();
        gtk_entry_set_activates_default (GTK_ENTRY (entry), TRUE);

        gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, FALSE, 0);
        gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
        gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
        gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox),
                            hbox, FALSE, FALSE, 0);

        gtk_widget_show_all (dialog);

        res = gtk_dialog_run (GTK_DIALOG (dialog));
        switch (res) {
        case GTK_RESPONSE_OK:
                *elname = g_strdup (gtk_entry_get_text (GTK_ENTRY (entry)));
                ret = TRUE;
                break;
        case GTK_RESPONSE_CANCEL:
        case GTK_RESPONSE_DELETE_EVENT:
		break;
        default:
                g_assert_not_reached ();
	}

        gtk_widget_destroy (dialog);
        return ret;
}


/**
 *Interactively create a new document.
 *Asks the user for info like 
 *root element, dtd (if validation is switched on) ...etc.
 *
 *@param a_this the current mlview editor.
 */
void
mlview_editor_create_new_xml_document (MlViewEditor * a_this)
{
        struct MlViewAppSettings *settings = NULL;
        MlViewXMLDocument *mlview_doc = NULL;
        xmlDocPtr xml_doc = NULL;
        xmlNodePtr xml_node = NULL;
        MlViewIView *view = NULL ;
        gboolean loop = TRUE;
        gchar *utf8_elname = NULL;

        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);
        g_return_if_fail (PRIVATE (a_this)->app_context != NULL);

        settings = mlview_app_context_get_settings 
                (PRIVATE (a_this)->app_context);
        g_return_if_fail (settings);

        while (loop) {
                gboolean res = FALSE;
                enum MlViewStatus status = MLVIEW_OK;
                guchar *name_end = NULL;

                res = mlview_editor_ask_root_element_name 
                        (&utf8_elname);

                /*the user hit cancel */
                if (!res)
                        break;
                /*empty root name: show the dialog again */
                if (!utf8_elname)
                        continue;

                if (mlview_utils_is_white_string (utf8_elname) == TRUE)
                        continue;

                /*check if is a valid element name */
                status = mlview_utils_parse_element_name (utf8_elname,
                                                          &name_end);
                if (status != MLVIEW_OK || !name_end) {
                        mlview_app_context_error (PRIVATE (a_this)->app_context,
                                _("The string entered is not a well formed element name !!"));
                        continue;
                }

                xml_node = xmlNewNode (NULL, utf8_elname);
                xml_doc = xmlNewDoc ("1.0");
                xml_doc->name = g_strdup ("untitled");
                xmlDocSetRootElement (xml_doc, xml_node);

                mlview_doc = mlview_xml_document_new (xml_doc,
                                                      PRIVATE (a_this)->
                                                      app_context);
                g_return_if_fail (mlview_doc != NULL);

                /*validation */
                if (settings->general.validation_is_on == TRUE) {
                        mlview_xml_document_associate_dtd_interactive
                                       (mlview_doc);
                        if (xml_node->type == XML_ELEMENT_NODE) {
                                mlview_parsing_utils_build_required_attributes_list
                                                (PRIVATE (a_this)->app_context,
                                                 xml_node);

                                mlview_parsing_utils_build_required_children_tree
                                                (PRIVATE
                                                 (a_this)->app_context,
                                                 &xml_node);
                        }
                }

                view = MLVIEW_IVIEW (mlview_tree_view_new (mlview_doc, 
                                            (gchar *) "", 
                                            PRIVATE (a_this)->app_context));
                                            mlview_editor_add_xml_document_view
                                                    (a_this, view);
                /*everything went fine */
                break;
        }

        if (utf8_elname) {
                g_free (utf8_elname);
                utf8_elname = NULL ;
        }
}


/**
 *Edits the xml document given in argument.
 *Actually, this method creates a view on this document and
 *adds the view to the editor.
 *@param a_this the current mlview editor.
 *@param a_doc the xml document to edit. This is an
 *object of the libxml2
 *@param a_doc_name the name of the xml document to edit.
 */
void
mlview_editor_edit_xml_document (MlViewEditor * a_this,
                                 xmlDocPtr a_doc,
                                 gchar * a_doc_name)
{
        MlViewIView *doc_view = NULL ;
        MlViewXMLDocument *mlview_xml_doc = NULL;

        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);
        g_return_if_fail (a_doc != NULL);

        mlview_xml_doc =
                mlview_xml_document_new
                (a_doc, PRIVATE (a_this)->app_context) ;

        doc_view =
                MLVIEW_IVIEW
                (mlview_tree_view_new
                 (mlview_xml_doc,
                  (guchar *) "",
                  PRIVATE (a_this)->app_context)) ;
        mlview_editor_add_xml_document_view (a_this, doc_view);
}

/**
 *Execute an editing editing action.
 *@param a_this the current instance of #MlViewEditor
 *@param a_action the action to execute.
 *@return MLVIEW_OK upon successful completion an error code
 *otherwise.
 */
enum MlViewStatus
mlview_editor_execute_action (MlViewEditor *a_this,
                              MlViewAction *a_action)
{
        enum MlViewStatus status = MLVIEW_OK ;

        g_return_val_if_fail (a_this
                              && MLVIEW_IS_EDITOR (a_this)
                              && PRIVATE (a_this),
                              MLVIEW_BAD_PARAM_ERROR) ;

        if (PRIVATE (a_this)->cur_view) {
                status = mlview_iview_execute_action
                        (PRIVATE (a_this)->cur_view,
                         a_action) ;
        }
        return status ;
}


/**
 *Getter of the file path of the document 
 *associated to the currently selected view.
 *@param a_this the current mlview editor.
 *@return the file path.
 */
gchar *
mlview_editor_get_current_xml_doc_file_path (MlViewEditor *a_this)
{
        MlViewXMLDocument *doc = NULL;

        g_return_val_if_fail (a_this != NULL, NULL);
        g_return_val_if_fail (MLVIEW_IS_EDITOR (a_this), NULL);
        g_return_val_if_fail (PRIVATE (a_this) != NULL, NULL);

        mlview_iview_get_document
                (PRIVATE (a_this)->cur_view, &doc);
        g_return_val_if_fail (doc != NULL, NULL);

        return mlview_xml_document_get_file_path (doc);
}


/**
 *Interactively saves the underlying xml document of the 
 *currently selected document view.
 *If the document has an associated file name, 
 *save it in that file or else, asks the user where to
 *save it.
 *@param a_this the current mlview editor.
 */
void
mlview_editor_save_xml_document (MlViewEditor * a_this)
{
        MlViewXMLDocument *xml_doc = NULL;
        gchar *file_path = NULL;

        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);
        if (!PRIVATE (a_this)->cur_view)
                return;

        mlview_iview_get_document (PRIVATE (a_this)->cur_view, &xml_doc) ;
        if (xml_doc == NULL)
                return;

        file_path =
                mlview_editor_get_current_xml_doc_file_path
                (a_this);

        if (file_path == NULL)
                mlview_editor_save_xml_document_as_interactive
                        (a_this);
        else
                mlview_editor_save_xml_document_as (a_this,
                                                    file_path);

}


/**
 *Classical "save as" functionnality.
 *Saves the underlying document of the currently selected view
 *into the file denoted by the file path a_file_path.
 *
 *@param a_this the current mlview editor.
 *@param a_file_path file path where to save the document.
 */
void
mlview_editor_save_xml_document_as (MlViewEditor * a_this,
                                    gchar * a_file_path)
{
        MlViewXMLDocument *mlview_xml_document = NULL;
        gboolean file_was_untitled = FALSE;
        guchar *prev_file_path = NULL;

        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);
        if (!PRIVATE (a_this)->cur_view)
                return;
        g_return_if_fail (PRIVATE (a_this)->opened_file_paths);
        g_return_if_fail (a_file_path != NULL);

        mlview_iview_get_document (PRIVATE (a_this)->cur_view,
                                   &mlview_xml_document) ;
        mlview_app_context_sbar_push_message
                (PRIVATE (a_this)->app_context,
                 _("Saving xml document as file %s..."),
                 a_file_path);

        file_was_untitled =
                (mlview_xml_document_get_file_descriptor (mlview_xml_document)
                 == NULL);

        prev_file_path =
                mlview_xml_document_get_file_path
                (mlview_xml_document);

        /*really save the document now */
        if (mlview_xml_document_save (mlview_xml_document,
                                      a_file_path, TRUE) > 0
            && (!prev_file_path
                || strcmp (prev_file_path, a_file_path))) {
                guchar *new_file_path = NULL;

                /*
                 *The save was OK and the new file path of this doc
                 * is different from the previous file path of this doc.
                 */

                /*
                 *remove the reference to the previous file path 
                 *of this document.
                 */
                if (prev_file_path) {
                        g_hash_table_remove
                                (PRIVATE (a_this)->
                                 opened_file_paths,
                                 prev_file_path);
                }

                /*
                 *Now, reference the new path of this doc.
                 *We must make sure that the string we put in the
                 *hash table belongs to mlview_xml_document.
                 *That way, one can destroy the hash table without
                 *any fear to leak the memory hold by the string.
                 */
                new_file_path = mlview_xml_document_get_file_path
                        (mlview_xml_document);

                if (new_file_path)
                        g_hash_table_insert (PRIVATE (a_this)->
                                             opened_file_paths,
                                             new_file_path,
                                             PRIVATE (a_this)->
                                             cur_view);
        }

        mlview_app_context_sbar_pop_message (PRIVATE (a_this)->
                                             app_context);
}


/**
 *Interactively saves the underlying document of the currently
 *selected view. Graphically asks the user where to save the
 *document and saves it.
 *
 *@param a_this the current mlview editor.
 */
void
mlview_editor_save_xml_document_as_interactive (MlViewEditor *a_this)
{
        MlViewFileSelection *file_sel;
        enum MLVIEW_SELECTED_BUTTON button;
        gchar *file_name = NULL;

        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);
        g_return_if_fail (PRIVATE (a_this)->app_context !=
                          NULL);

        file_sel =
                mlview_app_context_get_file_selector
                (PRIVATE (a_this)->app_context,
                 _("Choose a xml document"));

        g_return_if_fail (file_sel);

        button = mlview_file_selection_run (file_sel, TRUE);

        switch (button) {
        case OK_BUTTON:
                file_name =
                        g_strdup (gtk_file_selection_get_filename
                                  (GTK_FILE_SELECTION
                                   (file_sel)));

                if (file_name && strcmp (file_name, "")) {
                        mlview_editor_save_xml_document_as
                                (a_this, file_name);
                }

                if (file_name) {
                        g_free (file_name);
                        file_name = NULL;
                }
                break;
        case CANCEL_BUTTON:
        case WINDOW_CLOSED:
        default:
                break;
        }
}

/**
 *closes the current view without saving the underlying document.
 *@param a_this the current mlview editor.
 */
void
mlview_editor_close_xml_document_without_saving (MlViewEditor *a_this)
{
        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);

        mlview_editor_remove_xml_document_view
                (a_this,
                 PRIVATE (a_this)->cur_view);

        if (g_hash_table_size
            (PRIVATE (a_this)->mlview_xml_docs) == 0)
                PRIVATE (a_this)->cur_view = NULL;
}


/**
 *Saves the underlying document of the currently selected view
 *and closes the view.
 *
 *@param a_this the current mlview editor.
 */
void
mlview_editor_save_and_close_xml_document (MlViewEditor *a_this)
{
        MlViewFileDescriptor *file_desc = NULL ;
        MlViewFileSelection *file_sel = NULL;
        MlViewXMLDocument *mlview_xml_document = NULL;
        enum MLVIEW_SELECTED_BUTTON button;
        gchar *file_name = NULL, *tmp_str = NULL;

        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);
        if (!PRIVATE (a_this)->cur_view)
                return;
        g_return_if_fail (PRIVATE (a_this)->app_context !=
                          NULL);
        
        mlview_iview_get_document (PRIVATE (a_this)->cur_view,
                                   &mlview_xml_document) ;
        g_return_if_fail (mlview_xml_document != NULL);

        file_desc = mlview_xml_document_get_file_descriptor 
                (mlview_xml_document) ;
        if (!file_desc) {
                /*
                 *No file is associated to the document.
                 *=>Ask the user which file she wants to save
                 *the document in.
                 */
                file_sel = mlview_app_context_get_file_selector
                        (PRIVATE (a_this)->app_context,
                         _("Choose a xml document"));
                g_return_if_fail (file_sel);

                button = mlview_file_selection_run (file_sel, TRUE);

                switch (button) {
                case OK_BUTTON:
                        file_name =
                                g_strdup (gtk_file_selection_get_filename
                                          (GTK_FILE_SELECTION
                                           (file_sel))) ;
                        break;
                case CANCEL_BUTTON:
                case WINDOW_CLOSED:
                        break;
                }
        } else {
                /*The document already has*/
                tmp_str = mlview_file_descriptor_get_file_path (file_desc);
                g_return_if_fail (tmp_str);
                file_name = g_strdup (tmp_str);
        }
        if (file_name && strcmp (file_name, "")) {
                mlview_xml_document_save
                        (mlview_xml_document, file_name,
                         TRUE);
                mlview_editor_close_xml_document_without_saving
                        (a_this);
        }
        if (file_name) {
                g_free (file_name);
                file_name = NULL;
        }
}


/**
 *Closes all the views (and their underlying documents).
 *Save the documents if needed before closing the views.
 *
 *@param a_this the current mlview editor.
 */
void
mlview_editor_close_all_xml_documents_interactive (MlViewEditor *a_this)
{
        GList *views = NULL,
                *mobile_view_ptr = NULL;

        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);
        g_return_if_fail (PRIVATE (a_this)->
                          mlview_xml_doc_views != NULL);

        views = build_view_list_from_hashtable
                (PRIVATE (a_this)->mlview_xml_doc_views);

        if (views == NULL)
                return;

        for (mobile_view_ptr = views;
             mobile_view_ptr;
             mobile_view_ptr = mobile_view_ptr->next) {
                PRIVATE (a_this)->cur_view =
                        (MlViewIView *) mobile_view_ptr->data;
                mlview_editor_close_xml_document_interactive (a_this);
        }
}


/**
 *Shows a confirmation dialog before closing a modified document.
 *
 *@param a_this the current mlview editor.
 */
static void
mlview_editor_confirm_close (MlViewEditor *a_this)
{
        GtkWidget *dialog;
        gint ret;

        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this));

        dialog = gtk_message_dialog_new (NULL,
                                         GTK_DIALOG_MODAL,
                                         GTK_MESSAGE_QUESTION,
                                         GTK_BUTTONS_NONE,
                                         _("The document has been modifed.\n"
					   "Should I save it before closing it?"));

        gtk_dialog_add_buttons (GTK_DIALOG (dialog),
                                _("_Close without Saving"), GTK_RESPONSE_NO,
                                GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                GTK_STOCK_SAVE, GTK_RESPONSE_YES, NULL);

        gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
        gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_YES);

        ret = gtk_dialog_run (GTK_DIALOG (dialog));
        switch (ret) {
        case GTK_RESPONSE_YES:
                mlview_editor_save_and_close_xml_document (a_this);
                break;
        case GTK_RESPONSE_NO:
                mlview_editor_close_xml_document_without_saving (a_this);
                break;
        case GTK_RESPONSE_CANCEL:
        case GTK_RESPONSE_DELETE_EVENT:
                break ;
        default:
                g_assert_not_reached ();
        }

        gtk_widget_destroy (dialog);
}


/**
 *Interactively closes the current document view.
 *Saves the underlying document if needed.
 *
 *@param a_this the current mlview editor.
 */
void
mlview_editor_close_xml_document_interactive (MlViewEditor *a_this)
{
        MlViewXMLDocument *doc = NULL;

        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this));

        if (!PRIVATE (a_this)->cur_view)
                return ;
        mlview_iview_get_document (PRIVATE (a_this)->cur_view,
                                   &doc) ;
        if (!doc) {
                mlview_utils_trace_info
                        ("The current view has no associated document. "
                         "This is truly weird, something bad is happening.") ;
                return ;
        }

        if (mlview_xml_document_needs_saving (doc) == FALSE)
                mlview_editor_close_xml_document_without_saving (a_this);
        else
                mlview_editor_confirm_close (a_this);
}

/**
 *Interactively edits the editor settings.
 *
 *@param a_this the current mlview editor.
 */
void
mlview_editor_edit_settings_interactive (MlViewEditor * a_this)
{
        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);

}


/**
 *Associates a dtd to the current mlview xml document. 
 *
 *@param a_this the current mlview editor.
 *@return 0 if association worked, 1 if not xml document is opened, 
 */
gint
mlview_editor_associate_dtd_interactive (MlViewEditor * a_this)
{

        MlViewXMLDocument *mlview_xml_doc = NULL;

        g_return_val_if_fail (a_this != NULL, -1);
        g_return_val_if_fail (MLVIEW_IS_EDITOR (a_this), -1);
        g_return_val_if_fail (PRIVATE (a_this) != NULL, -1);

        if (!PRIVATE (a_this)->cur_view)
                return 1;

        mlview_iview_get_document (PRIVATE (a_this)->cur_view,
                                   &mlview_xml_doc) ;
        if (mlview_xml_doc == NULL)
                return 1;

        if (mlview_xml_document_associate_dtd_interactive
            (mlview_xml_doc))
                return 0;

        return 2;
}

/**
 *validate the current selected document against the dtd it is associated to.
 *If there is no dtd associated to the document, the functions asks the user
 *to associate a dtd to it and then, does the validation 
 *
 *@param a_this the current mlview editor.
 */
gint
mlview_editor_validate (MlViewEditor * a_this)
{

        MlViewXMLDocument *doc = NULL;

        g_return_val_if_fail (a_this != NULL, -1);
        g_return_val_if_fail (MLVIEW_IS_EDITOR (a_this), -1);
        g_return_val_if_fail (PRIVATE (a_this) != NULL, -1);

        if (!PRIVATE (a_this)->cur_view)
                return 1;

        mlview_iview_get_document (PRIVATE (a_this)->cur_view,
                                   &doc) ;

        if (doc == NULL)
                return 1;

        return mlview_xml_document_validate (doc);
}


/**
 *Applies a XSLT stylesheet to the current document
 *displaying a dialog for xslt selection and opens
 *the transformation resulting document in the editor
 *@param a_this the current instance of #MlViewEditor.
 *
 */
void
mlview_editor_xslt_transform_document_interactive (MlViewEditor *a_this)
{
        MlViewXMLDocument *src_doc = NULL;
        MlViewXMLDocument *xsl_doc = NULL;
        MlViewXMLDocument *res_doc = NULL;

        g_return_if_fail (a_this != NULL);
        g_return_if_fail (MLVIEW_IS_EDITOR (a_this));
        g_return_if_fail (PRIVATE (a_this) != NULL);
        
        src_doc = mlview_editor_get_cur_doc (a_this);
        xsl_doc = mlview_xslt_utils_select_xsl_doc (a_this);
        if (xsl_doc != NULL) {
                res_doc = mlview_xslt_utils_transform_document
                        (src_doc, xsl_doc);
                mlview_xml_document_unref(xsl_doc);
                if (res_doc != NULL) 
                        mlview_editor_create_new_view_on_document (a_this, res_doc, 
                                                                   MLVIEW_XML_DOC_TREE_VIEW);
        }
}
